#!/usr/bin/env python2.5
# -*- coding: us-ascii -*-
###############################################################################
# tinyurl.py - A tinyurl implementation in Python
# Copyright (C) 2006  Dwayne C. Litzenberger <dlitz@dlitz.net>
###############################################################################
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
###############################################################################
# Version 0.1  2006-10-22  Dwayne C. Litzenberger  <dlitz@dlitz.net>
#  * Initial release
###############################################################################

# Probably requires Python >= 2.5 

import string
import sqlite3
import re
import cgi
from wsgiref.handlers import CGIHandler
from wsgiref.util import application_uri

assert sqlite3.apilevel == "2.0"
#assert sqlite3.threadsafety >= 0
assert sqlite3.paramstyle == "qmark"

class TinyUrlApp(object):

    ## CONFIGURATION ##
    database_path = "/var/www/tinyurl/user/db/tinyurldb"
    base_url = None
    ## END CONFIGURATION ##
    
    # NB: Updating allowed chars will break number2id/id2number
    # NB: allowed_chars must be regexp-safe, url-safe, and filename-safe
    allowed_chars = 'acefjkmnpqrsuxy01234689'
    
    trans_table = string.maketrans("odli7tbgh5wvz", "0a111f69nsuu2")

    assert allowed_chars == re.escape(allowed_chars)
    illegal_chars_regexp = re.compile(r'[^'+re.escape(allowed_chars)+']')

    def __init__(self, environ, start_response):
        self.environ = environ
        self.start_response = start_response
        self.db_conn = None

    def __iter__(self):
        self.db_connect()
        self.create_tables()

        request_method = self.environ.get('REQUEST_METHOD', 'GET')
        path_info = self.environ.get('PATH_INFO', '')
        id = self.normalize_id(path_info[1:])

        if id:
            url = self.lookup(id)
            if url:
                status = "301 Moved Permanently"
                response_headers = [('Content-type', 'text/html; charset=US-ASCII'), ('Location', url)]
                self.start_response(status, response_headers)
                yield self.get_html('<p>Redirecting to <a href="%(url_html)s">%(url_html)s</a></p>' % { 'url_html': cgi.escape(url) })
                return
            else:
                status = "404 Not Found"
                response_headers = [('Content-type', 'text/html; charset=US-ASCII')]
                self.start_response(status, response_headers)
                yield self.get_html('<p>No such URL</p>')
                return

        if request_method == "POST":
            form = cgi.FieldStorage(fp=self.environ['wsgi.input'], environ=self.environ)
            longurl = form.getvalue('longurl', '').strip()
            if longurl and form.getvalue('action_create_tinyurl'):

                id = self.insert(longurl)
                if self.base_url:
                    shorturl = self.base_url.rstrip("/") + "/" + id
                else:
                    shorturl = application_uri(self.environ).rstrip("/") + "/" + id
                status = '200 OK'
                response_headers = [('Content-type', 'text/html; charset=US-ASCII')]
                self.start_response(status, response_headers)
                yield self.get_inserted_html(shorturl=shorturl, longurl=longurl)
                return

        status = '200 OK'
        response_headers = [('Content-type', 'text/html; charset=US-ASCII')]
        self.start_response(status, response_headers)
        yield self.get_form_html()

    def db_connect(self):
        if self.db_conn is None:
            self.db_conn = sqlite3.connect(self.database_path)

    def create_tables(self):
        c = self.db_conn.cursor()

        # Create the urls table
        c.execute('''
            CREATE TABLE IF NOT EXISTS urls (
                number  INTEGER NOT NULL PRIMARY KEY,
                url     TEXT  NOT NULL UNIQUE
            )
            ''')

    def lookup(self, id):
        c = self.db_conn.cursor()
        try:
            number = self.id2number(id)
        except ValueError:
            return None
        c.execute('''SELECT url FROM urls WHERE number = ?''', (number,))
        row = c.fetchone()
        if not row:
            return None
        return row[0].encode('US-ASCII')

    def insert(self, url):
        c = self.db_conn.cursor()
        
        # Insert the row into the database
        c.execute('''INSERT OR IGNORE INTO urls (url) VALUES (?)''', (unicode(url, 'US-ASCII'),))
        self.db_conn.commit()
        
        c.execute('''SELECT number FROM urls WHERE url = ?''', (unicode(url, 'US-ASCII'),))
        row = c.fetchone()
        if not row:
            raise RuntimeError, "unable to insert url into database"
        number = row[0]
        return self.number2id(number)
 
    def number2id(self, n):
        n = int(n)
        if n < 0:
            raise ValueError, "n must not be negative"
        if n == 0:
            return self.allowed_chars[0]
        modulo = len(self.allowed_chars)
        retval = []
        while n > 0:
            n, i = divmod(n, modulo)
            retval.insert(0, self.allowed_chars[i])
        return "".join(retval)

    def id2number(self, id):
        id = self.normalize_id(id)
        if len(id) < 1:
            raise ValueError, "id must not be empty"
        if id == self.allowed_chars[0]:
            return 0
        modulo = len(self.allowed_chars)
        retval = 0
        for ch in id:
            i = self.allowed_chars.index(ch)
            retval = (retval * modulo) + i
        return retval

    def normalize_id(self, s):
        return self.illegal_chars_regexp.sub("", s.lower().translate(self.trans_table))
    
    def get_html(self, body_html='', title_html=''):
        return """<!DOCTYPE html
  PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
  "http://www.w3.org/TR/xhtml1/DTD/xhtml1.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">

<head>

  <title>%(title_html)s</title>

</head>

<body>
%(body_html)s

    <p><a href="http://ftp.dlitz.net/pub/dlitz/tinyurl/">Source code</a></p>

</body>

</html>
""" % { 'title_html': title_html, 'body_html': body_html }

    def get_form_html(self):
    
        return self.get_html(body_html='''
    <form action="" method="post">
        <div>
            <label for="longurl">Long URL: <input type="text" name="longurl" id="longurl" /></label><br />
            <input type="submit" name="action_create_tinyurl" value="Create Short URL" />
        </div>
    </form>

''')
    
    def get_inserted_html(self, shorturl, longurl):
        return self.get_html(body_html='''
    <p>The following URL:</p>
    
    <blockquote><p><b>%(longurl_html)s</b></p></blockquote>

    <p>is now referenced by the URL:</p>

    <blockquote><p><b><a href="%(shorturl_html)s">%(shorturl_html)s</a></b></p></blockquote>
    
    <hr />

    <form action="" method="post">
        <div>
            <label for="longurl">Long URL: <input type="text" name="longurl" id="longurl" /></label><br />
            <input type="submit" name="action_create_tinyurl" value="Create Short URL" />
        </div>
    </form>

''' % {'shorturl_html': cgi.escape(shorturl), 'longurl_html': cgi.escape(longurl)})

 
if __name__ == '__main__':
    CGIHandler().run(TinyUrlApp)
